import numpy as np
import numpy.linalg as npla


def pol2cart(pts: np.ndarray) -> np.ndarray:
  """Converts points from polar to cartesian.
  Args:
    pts (np.ndarray): points in polar coordinate [rho, theta, phi]
  Returns:
    np.ndarray: points in cartesian coordinate [x, y, z]
  """
  assert pts.shape[-2:] == (3, 1)
  x = np.cos(pts[..., 2:3, :]) * np.sin(pts[..., 1:2, :]) * pts[..., 0:1, :]
  y = np.sin(pts[..., 2:3, :]) * np.sin(pts[..., 1:2, :]) * pts[..., 0:1, :]
  z = np.cos(pts[..., 1:2, :]) * pts[..., 0:1, :]
  return np.concatenate((x, y, z), axis=-2)


def cart2pol(pts: np.ndarray) -> np.ndarray:
  """Converts points from cartesian to polar.
  Args:
    pts (np.ndarray): points in cartesian coordinate [x, y, z]
  Returns:
    np.ndarray: points in polar coordinate [rho, theta, phi]
  """
  assert pts.shape[-2:] == (3, 1)
  rho = npla.norm(pts, axis=-2, keepdims=True)
  phi = np.arctan2(pts[..., 1:2, :], pts[..., 0:1, :])
  theta = np.arctan2(np.sqrt(pts[..., 0:1, :]**2 + pts[..., 1:2, :]**2), pts[..., 2:3, :])
  return np.concatenate((rho, theta, phi), axis=-2)


from __future__ import annotations

from typing import Optional

import numpy as np

from . import operations as so3op


class Rotation:
  """Light weight rotation class."""

  def __init__(self,
               *,
               rotation: Optional[Rotation] = None,
               C_ba: Optional[np.ndarray] = None,
               aaxis_ab: Optional[np.ndarray] = None,
               num_terms: int = 0) -> None:
    """Constructor with several options.
    Args:
      rotation (Optional[Rotation]): an existing instance of this class. (copy construction)
      C_ba (Optional[np.ndarray]): constructs from a 3x3 rotation matrix
      aaxis_ab (Optional[np.ndarray]): constructs using the exponential map from a 3x1 axis-angle vector
      num_terms (int): number of terms in the infintie series when converting aaxis_ab to C_ba; defaults to 0
        (analytical)
    """

    # copy construction
    if rotation is not None:
      self._C_ba = np.copy(rotation.matrix())
      return

    # constructs from a 3x3 rotation matrix
    if C_ba is not None:
      self._C_ba = np.copy(C_ba)
      self.reproject()
      return

    # construct using the exponential map from a 3x1 axis-angle vector
    if aaxis_ab is not None:
      assert aaxis_ab.shape[-2:] == (3, 1), "aaxis_ab has invalid shape."
      self._C_ba = so3op.vec2rot(aaxis_ab, num_terms)
      return

    # default construction
    self._C_ba = np.eye(3)

  def matrix(self) -> np.ndarray:
    """Returns the underlying rotation matrix representation."""
    return self._C_ba

  def vec(self) -> np.ndarray:
    """Returns the corresponding Lie algebra using the logarithmic map."""
    return so3op.rot2vec(self._C_ba)

  def inverse(self) -> Rotation:
    """Returns the inverse rotation."""
    temp = Rotation(C_ba=self._C_ba.swapaxes(-2, -1))
    temp.reproject()
    return temp

  def reproject(self) -> None:
    """Reprojects the rotation matrix back onto SO(3)."""
    self._C_ba = so3op.vec2rot(so3op.rot2vec(self._C_ba))

  def __matmul__(self, other: Rotation) -> Rotation:
    return Rotation(C_ba=self.matrix() @ other.matrix())

  def __mul__(self, other: Rotation) -> Rotation:
    return Rotation(C_ba=self.matrix() @ other.matrix())

  def __str__(self):
    return str(self.matrix())


from .common import operations as cmnop
from .so3 import operations as so3op
from .se3 import operations as se3op
from .so3.rotation import Rotation
from .se3.transformation import Transformation
from .se3.transformation_with_covariance import TransformationWithCovariance


from __future__ import annotations

from typing import Optional

import numpy as np

from ..so3 import operations as so3op
from . import operations as se3op


class Transformation:
  """Light weight transformation class."""

  def __init__(self,
               *,
               transformation: Optional[Transformation] = None,
               T_ba: Optional[np.ndarray] = None,
               C_ba: Optional[np.ndarray] = None,
               r_ba_in_a: Optional[np.ndarray] = None,
               xi_ab: Optional[np.ndarray] = None,
               num_terms: int = 0) -> None:
    """Constructor with several options.
    Args:
      transformation (Optional[Transformation]): an existing instance of this class. (copy construction)
      T_ba (Optional[np.ndarray]): constructs from a 4x4 transformation matrix
      C_ba (Optional[np.ndarray]): constructs from a 3x3 rotation matrix and a 3x1 translation vector
      r_ba_in_a (Optional[np.ndarray]): constructs from a 3x3 rotation matrix and a 3x1 translation vector
      xi_ab (Optional[np.ndarray]): constructs using the exponential map from a 6x1 se3 algebra vector
      num_terms (int): number of terms in the infintie series when converting xi_ab to T_ba; defaults to 0 (analytical)
    """

    # copy construction
    if transformation is not None:
      self._C_ba = np.copy(transformation.C_ba())
      self._r_ab_inb = np.copy(transformation.r_ab_inb())
      return

    # constructs from a 4x4 transformation matrix
    if T_ba is not None:
      self._C_ba = np.copy(T_ba[..., 0:3, 0:3])
      self._r_ab_inb = np.copy(T_ba[..., 0:3, 3:4])
      return

    # constructs from a 3x3 rotation matrix and a 3x1 translation vector
    if (C_ba is not None) and (r_ba_in_a is not None):
      assert C_ba.shape[:-2] == r_ba_in_a.shape[:-2]
      self._C_ba = np.copy(C_ba)
      self._r_ab_inb = -C_ba @ r_ba_in_a
      self.reproject()
      return

    # constructs using the exponential map from a 6x1 se3 algebra vector
    if xi_ab is not None:
      assert xi_ab.shape[-2:] == (6, 1), "xi_ab has invalid shape."
      self._C_ba, self._r_ab_inb = se3op.vec2tran(xi_ab, num_terms, False)
      return

    # default construction
    self._C_ba: np.ndarray = np.eye(3)
    self._r_ab_inb: np.ndarray = np.zeros((3, 1))

  def assign(self,
             *,
             T_ba: Optional[np.ndarray] = None,
             C_ba: Optional[np.ndarray] = None,
             r_ba_in_a: Optional[np.ndarray] = None,
             xi_ab: Optional[np.ndarray] = None,
             num_terms: int = 0) -> None:
    """Assigns a new transformation to this class. (operator= in c++)
    Args:
      T_ba (Optional[np.ndarray]): constructs from a 4x4 transformation matrix
      C_ba (Optional[np.ndarray]): constructs from a 3x3 rotation matrix and a 3x1 translation vector
      r_ba_in_a (Optional[np.ndarray]): constructs from a 3x3 rotation matrix and a 3x1 translation vector
      xi_ab (Optional[np.ndarray]): constructs using the exponential map from a 6x1 se3 algebra vector
      num_terms (int): number of terms in the infintie series when converting xi_ab to T_ba; defaults to 0 (analytical)
    """

    if T_ba is not None:
      self._C_ba[:] = T_ba[..., 0:3, 0:3]
      self._r_ab_inb[:] = T_ba[..., 0:3, 3:4]
      return

    if (C_ba is not None) and (r_ba_in_a is not None):
      assert C_ba.shape[:-2] == r_ba_in_a.shape[:-2]
      self._C_ba[:] = C_ba
      self._r_ab_inb[:] = -C_ba @ r_ba_in_a
      self.reproject()
      return

    if xi_ab is not None:
      assert xi_ab.shape[-2:] == (6, 1), "xi_ab has invalid shape."
      self._C_ba[:], self._r_ab_inb[:] = se3op.vec2tran(xi_ab, num_terms, False)
      return

  def matrix(self) -> np.ndarray:
    """Returns the transformation matrix representation."""
    T_ba = np.zeros(self._C_ba.shape[:-2] + (4, 4))
    T_ba[..., :3, :3] = self._C_ba
    T_ba[..., :3, 3:4] = self._r_ab_inb
    T_ba[..., 3, 3] = 1
    return T_ba

  def C_ba(self) -> np.ndarray:
    """Returns the underlying rotation matrix."""
    return self._C_ba

  def r_ba_ina(self) -> np.ndarray:
    """Returns the "forward" translation r_ba_ina = -C_ba.transpose()*r_ab_inb."""
    return -self._C_ba.swapaxes(-2, -1) @ self._r_ab_inb

  def r_ab_inb(self) -> np.ndarray:
    """Returns the underlying r_ab_inb vector."""
    return self._r_ab_inb

  def vec(self) -> np.ndarray:
    """Returns the corresponding Lie algebra using the logarithmic map."""
    return se3op.tran2vec(self._C_ba, self._r_ab_inb)

  def inverse(self) -> Transformation:
    """Returns the inverse transformation."""
    temp = Transformation(C_ba=self._C_ba.swapaxes(-2, -1), r_ba_in_a=self._r_ab_inb)
    temp.reproject()
    return temp

  def adjoint(self) -> np.ndarray:
    """Returns the 6x6 adjoint transformation matrix."""
    return se3op.tranAd(self._C_ba, self._r_ab_inb)

  def reproject(self) -> None:
    """Reprojects the transformation matrix back onto SE(3)."""
    self._C_ba = so3op.vec2rot(so3op.rot2vec(self._C_ba))

  def __matmul__(self, other) -> Transformation:
    return Transformation(T_ba=self.matrix() @ other.matrix())

  def __mul__(self, other) -> Transformation:
    return Transformation(T_ba=self.matrix() @ other.matrix())

  def __str__(self):
    return str(self.matrix())


from typing import Optional

import numpy as np
import numpy.linalg as npla

from ..so3 import operations as so3op


def hat(xi_or_rho: np.ndarray, aaxis: Optional[np.ndarray] = None) -> np.ndarray:
  """Builds the 4x4 "skew symmetric matrix".

  The hat (^) operator, builds the 4x4 skew symmetric matrix from the 3x1 axis angle vector and 3x1 translation vector.
    hat(rho, aaxis) = [aaxis^ rho] = [0.0  -a3   a2  rho1]
                      [  0^T    0]   [ a3  0.0  -a1  rho2]
                                     [-a2   a1  0.0  rho3]
                                     [0.0  0.0  0.0   0.0]
  See eq. 4 in Barfoot-TRO-2014 for more information.

  Args:
    xi_or_rho (np.ndarray): either 6x1 se3 algebra vector xi, or 3x1 translation vector rho when aaxis is None
    aaxis (np.ndarray|None): 3x1 axis angle vector
  Returns:
    np.ndarray: the 4x4 "skew symmetric matrix"
  """
  if aaxis is None:
    assert xi_or_rho.shape[-2:] == (6, 1)
    rho = xi_or_rho[..., :3, :]
    aaxis = xi_or_rho[..., 3:, :]
  else:
    assert xi_or_rho.shape[-2:] == (3, 1)
    assert aaxis.shape[-2:] == (3, 1)
    rho = xi_or_rho

  x_hat = np.zeros(rho.shape[:-2] + (4, 4))
  x_hat[..., :3, :3] = so3op.hat(aaxis)
  x_hat[..., :3, 3:4] = rho
  return x_hat


def curlyhat(xi_or_rho, aaxis=None) -> np.ndarray:
  """Builds the 6x6 "curly hat" matrix (related to the skew symmetric matrix).

  The curly hat operator builds the 6x6 skew symmetric matrix from the 3x1 axis angle vector and 3x1 translation vector.
    curlyhat(rho, aaxis) = [aaxis^   rho^]
                           [     0 aaxis^]
  See eq. 12 in Barfoot-TRO-2014 for more information.

  Args:
    xi_or_rho (np.ndarray): either 6x1 se3 algebra vector xi, or 3x1 translation vector rho when aaxis is None
    aaxis (np.ndarray|None): 3x1 axis angle vector
  Returns:
    np.ndarray: the 6x6 "curly hat" matrix
  """
  if aaxis is None:
    assert xi_or_rho.shape[-2:] == (6, 1)
    rho = xi_or_rho[..., :3, :]
    aaxis = xi_or_rho[..., 3:, :]
  else:
    assert xi_or_rho.shape[-2:] == (3, 1)
    assert aaxis.shape[-2:] == (3, 1)
    rho = xi_or_rho

  x_curlyhat = np.zeros(rho.shape[:-2] + (6, 6))
  x_curlyhat[..., :3, :3] = x_curlyhat[..., 3:, 3:] = so3op.hat(aaxis)
  x_curlyhat[..., :3, 3:] = so3op.hat(rho)

  return x_curlyhat


def Cr2T(*, C_ab=None, r_ba_ina=None, r_ab_inb=None):
  assert C_ab is not None
  assert (r_ba_ina is not None) or (r_ab_inb is not None)
  T_ab = np.tile(np.eye(4), (*C_ab.shape[:-2], 1, 1))
  T_ab[..., 0:3, 0:3] = C_ab
  if r_ba_ina is not None:
    T_ab[..., 0:3, 3:4] = r_ba_ina
  else:
    T_ab[..., 0:3, 3:4] = -C_ab @ r_ab_inb
  return T_ab


def T2Cr(*, T_ab=None, return_r_ab_in_b=False):
  assert T_ab is not None
  C_ab = T_ab[..., 0:3, 0:3]
  r_ba_ina = T_ab[..., 0:3, 3:4]
  if return_r_ab_in_b:
    return C_ab, -np.swapaxes(C_ab, -2, -1) @ r_ba_ina
  else:
    return C_ab, r_ba_ina


def validate_tran(T):
  """causes an error if the transformation matrix is not in SE(3)"""
  so3op.validate_rot(T[..., 0:3, 0:3])
  err = np.max(np.abs(np.array([0., 0., 0., 1.]) - T[..., 3, :]))
  if err > 1e-10:
    raise RuntimeError('The bottom row of the transformation matrix should be [0,0,0,1]. Maximum error: {}'.format(err))


def _vec2tran_analytical(rho_ba, aaxis_ba):

  C_ab = so3op.vec2rot(aaxis_ba)
  J_ab = so3op.vec2jac(aaxis_ba)
  r_ba_ina = J_ab @ rho_ba

  T_ab = np.tile(np.eye(4), (*rho_ba.shape[:-2], 1, 1))
  T_ab[..., 0:3, 0:3] = C_ab
  T_ab[..., 0:3, 3:4] = r_ba_ina

  return T_ab


def _vec2tran_numerical(rho_ba, aaxis_ba, num_terms):

  T_ab = np.tile(np.eye(4), (*aaxis_ba.shape[:-2], 1, 1))

  xi_ba = np.concatenate((rho_ba, aaxis_ba), axis=-2)

  x_small = hat(xi_ba)
  x_small_n = np.tile(np.eye(4), (*aaxis_ba.shape[:-2], 1, 1))

  for n in range(1, num_terms + 1):
    x_small_n = x_small_n @ (x_small / n)
    T_ab = T_ab + x_small_n

  return T_ab


def _vec2tran_T(xi_ba, num_terms=0):
  if num_terms != 0:
    return _vec2tran_numerical(xi_ba[..., :3, :], xi_ba[..., 3:, :], num_terms)
  else:
    return _vec2tran_analytical(xi_ba[..., :3, :], xi_ba[..., 3:, :])


_vec2tran_T_vec = np.vectorize(_vec2tran_T, signature='(6,1),()->(4,4)')


def vec2tran(xi_ba: np.ndarray, num_terms: int = 0, return_T: bool = True) -> np.ndarray:
  """Builds a transformation matrix using the analytical exponential map or infinite series evaluation of the
  exponential map.

  This function builds a transformation matrix, T_ab, using the analytical exponential map, from the se3 algebra vector,
  xi_ba,
    T_ab = exp(xi_ba^) = [ C_ab r_ba_ina],   xi_ba = [  rho_ba]
                         [  0^T        1]            [aaxis_ba]
  where C_ab is a 3x3 rotation matrix from 'b' to 'a', r_ba_ina is the 3x1 translation vector from 'a' to 'b' expressed
  in frame 'a', aaxis_ba is a 3x1 axis-angle vector, the magnitude of the angle of rotation can be recovered by finding
  the norm of the vector, and the axis of rotation is the unit-length vector that arises from normalization.
  Note that the angle around the axis, aaxis_ba, is a right-hand-rule (counter-clockwise positive) angle from 'a' to
  'b'. The parameter, rho_ba, is a special translation-like parameter related to 'twist' theory. It is most inuitively
  described as being like a constant linear velocity (expressed in the smoothly-moving frame) for a fixed duration;
  for example, consider the curve of a car driving 'x' meters while turning at a rate of 'y' rad/s.
  For more information see Barfoot-TRO-2014 Appendix B1.
  Alternatively, we that note that
    T_ba = exp(-xi_ba^) = exp(xi_ab^).
  Both the analytical (num_terms = 0) or the numerical (num_terms > 0) may be evaluated.

  Args:
    xi_ba (np.ndarray): 6x1 se3 algebra vector xi
    num_terms (int): number of terms in the infinite series; use analytical solution of 0
    return_T (bool): returns T_ab if True; otherwise returns C_ab and r_ba_ina
  Returns:
    np.ndarray: T_ab
    or
    Tuple[np.ndarray, np.ndarray]: C_ab and r_ba_ina
  """
  assert xi_ba.shape[-2:] == (6, 1)
  T_ab = _vec2tran_T_vec(xi_ba, num_terms)
  return T_ab if return_T else T2Cr(T_ab=T_ab)


def tran2vec(T_or_C_ab: np.ndarray, r_ba_ina: Optional[np.ndarray] = None) -> np.ndarray:
  """Compute the matrix log of a transformation matrix.

  Compute the inverse of the exponential map (the logarithmic map). This lets us go from a 4x4 transformation matrix
  back to a 6x1 se3 algebra vector (composed of a 3x1 axis-angle vector and 3x1 twist-translation vector). In some
  cases, when the rotation in the transformation matrix is 'numerically off', this involves some 'projection' back to
  SE(3).
    xi_ba = ln(T_ab)
  where xi_ba is the 6x1 se3 algebra vector. Alternatively, we that note that
    xi_ab = -xi_ba = ln(T_ba) = ln(T_ab^{-1})
  See Barfoot-TRO-2014 Appendix B2 for more information.

  Args:
    T_or_C_ab (np.ndarray): either 4x4 transformation matrix T_ab, or 3x3 rotation matrix C_ab if r_ba_ina is None
    r_ba_ina (np.ndarray|None): 3x1 translation vector
  Returns:
    np.ndarray: 6x1 se3 algebra vector xi_ba
  """
  if r_ba_ina is None:
    assert T_or_C_ab.shape[-2:] == (4, 4)
    C_ab, r_ba_ina = T2Cr(T_ab=T_or_C_ab)
  else:
    assert T_or_C_ab.shape[-2:] == (3, 3)
    assert r_ba_ina.shape[-2:] == (3, 1)
    C_ab = T_or_C_ab

  aaxis_ba = so3op.rot2vec(C_ab)
  rho_ba = so3op.vec2jacinv(aaxis_ba) @ r_ba_ina

  xi_ba = np.concatenate((rho_ba, aaxis_ba), axis=-2)

  return xi_ba


def tranAd(T_or_C_ab, r_ba_ina=None) -> np.ndarray:
  """Builds the 6x6 adjoint transformation matrix from either the 3x3 rotation matrix and 3x1 translation vector, or the
  4x4 transformation matrix.
  Adjoint(T_ab) = Adjoint([C_ab r_ba_ina]) = [C_ab r_ba_ina^*C_ab] = exp(curlyhat(xi_ba))
                          ([ 0^T        1])   [   0           C_ab]
  See eq. 101 in Barfoot-TRO-2014 for more information.

  Args:
    T_or_C_ab (np.ndarray): either 4x4 transformation matrix T_ab, or 3x3 rotation matrix C_ab if r_ba_ina is None
    r_ba_ina (np.ndarray|None): 3x1 translation vector
  Returns:
    np.ndarray: the 6x6 adjoint transformation matrix
  """
  if r_ba_ina is None:
    C_ab, r_ba_ina = T2Cr(T_ab=T_or_C_ab)
  else:
    C_ab = T_or_C_ab

  adjoint_T_ab = np.zeros(T_or_C_ab.shape[:-2] + (6, 6))
  adjoint_T_ab[..., :3, :3] = C_ab
  adjoint_T_ab[..., 3:, 3:] = C_ab
  adjoint_T_ab[..., :3, 3:] = so3op.hat(r_ba_ina) @ C_ab

  return adjoint_T_ab


def _vec2Q(rho_ba, aaxis_ba):

  ang = npla.norm(aaxis_ba, axis=-2, keepdims=True)
  ang2 = ang * ang
  ang3 = ang2 * ang
  ang4 = ang3 * ang
  ang5 = ang4 * ang
  cang = np.cos(ang)
  sang = np.sin(ang)
  m2 = (ang - sang) / ang3
  m3 = (1.0 - 0.5 * ang2 - cang) / ang4
  m4 = 0.5 * (m3 - 3 * (ang - sang - ang3 / 6) / ang5)

  # construct matrix terms
  rx = so3op.hat(rho_ba)
  px = so3op.hat(aaxis_ba)
  pxrx = px @ rx
  rxpx = rx @ px
  pxrxpx = pxrx @ px

  # construct Q matrix
  return (0.5 * rx + m2 * (pxrx + rxpx + pxrxpx) - m3 * (px @ pxrx + rxpx @ px - 3 * pxrxpx) - m4 *
          (pxrxpx @ px + px @ pxrxpx))


def vec2Q(xi_or_rho_ba: np.ndarray, aaxis_ba: Optional[np.ndarray] = None) -> np.ndarray:
  """Constructs the 3x3 "Q" matrix, used in the 6x6 Jacobian of SE(3).

  See eqn 7.86a in Barfoot's book

  Args:
    xi_or_rho_ba (np.ndarray): either 6x1 se3 algebra vector xi, or 3x1 translation vector rho when aaxis is None
    aaxis_ba (np.ndarray|None): 3x1 axis angle vector
  Returns:
    np.ndarray: the 3x3 "Q" matrix
  """
  if aaxis_ba is None:
    assert xi_or_rho_ba.shape[-2:] == (6, 1)
    rho_ba = xi_or_rho_ba[..., :3, :]
    aaxis_ba = xi_or_rho_ba[..., 3:, :]
  else:
    assert xi_or_rho_ba.shape[-2:] == (3, 1)
    assert aaxis_ba.shape[-2:] == (3, 1)
    rho_ba = xi_or_rho_ba

  return _vec2Q(rho_ba, aaxis_ba)


def _vec2jac_analytical(rho_ba, aaxis_ba):

  J_ab = np.zeros(rho_ba.shape[:-2] + (6, 6))
  J_ab[..., :3, :3] = so3op.vec2jac(aaxis_ba)
  J_ab[..., 3:, 3:] = so3op.vec2jac(aaxis_ba)
  J_ab[..., :3, 3:] = _vec2Q(rho_ba, aaxis_ba)

  return J_ab


def _vec2jac_numerical(rho_ba, aaxis_ba, num_terms):
  J_ab = np.tile(np.eye(6), (*rho_ba.shape[:-2], 1, 1))

  xi_ba = np.concatenate((rho_ba, aaxis_ba), axis=-2)
  x_small = curlyhat(xi_ba)
  x_small_n = np.tile(np.eye(6), (*rho_ba.shape[:-2], 1, 1))

  for n in range(1, num_terms + 1):
    x_small_n = x_small_n @ (x_small / (n + 1))
    J_ab = J_ab + x_small_n

  return J_ab


def _vec2jac_rho_aaxis(rho_ba, aaxis_ba, num_terms=0):
  tolerance = 1e-12
  if (npla.norm(aaxis_ba, axis=-2)[0] < tolerance) or (num_terms != 0):
    return _vec2jac_numerical(rho_ba, aaxis_ba, num_terms)
  else:
    return _vec2jac_analytical(rho_ba, aaxis_ba)


_vec2jac_rho_aaxis_vec = np.vectorize(_vec2jac_rho_aaxis, signature='(3,1),(3,1),()->(6,6)')


def vec2jac(xi_or_rho_ba: np.ndarray, aaxis_ba: Optional[np.ndarray] = None, num_terms: int = 0) -> np.ndarray:
  """Build the 6x6 left Jacobian of SE(3).

  For the sake of a notation, we assign subscripts consistence with the transformation,
    J_ab = J(xi_ba)
  Where applicable, we also note that
    J(xi_ba) = Adjoint(exp(xi_ba^)) * J(-xi_ba),
  and
    Adjoint(exp(xi_ba^)) = identity + curlyhat(xi_ba) * J(xi_ba).
  For more information see eq. 100 in Barfoot-TRO-2014.

  Args:
    xi_or_rho_ba (np.ndarray): either 6x1 se3 algebra vector xi, or 3x1 translation vector rho when aaxis is None
    aaxis_ba (np.ndarray|None): 3x1 axis angle vector
    num_terms (int): number of terms used in the infinite series; use analytical solution if 0
  Returns:
    np.ndarray: the 6x6 left Jacobian of SE(3)
  """
  if aaxis_ba is None:
    assert xi_or_rho_ba.shape[-2:] == (6, 1)
    rho_ba = xi_or_rho_ba[..., :3, :]
    aaxis_ba = xi_or_rho_ba[..., 3:, :]
  else:
    assert xi_or_rho_ba.shape[-2:] == (3, 1)
    assert aaxis_ba.shape[-2:] == (3, 1)
    rho_ba = xi_or_rho_ba

  return _vec2jac_rho_aaxis_vec(rho_ba, aaxis_ba, num_terms)


def _vec2jacinv_analytical(rho_ba, aaxis_ba):

  J_ab_inv = np.zeros(rho_ba.shape[:-2] + (6, 6))
  J33_ab_inv = so3op.vec2jacinv(aaxis_ba)
  J_ab_inv[..., :3, :3] = J33_ab_inv
  J_ab_inv[..., 3:, 3:] = J33_ab_inv
  J_ab_inv[..., :3, 3:] = -J33_ab_inv @ _vec2Q(rho_ba, aaxis_ba) @ J33_ab_inv

  return J_ab_inv


def _vec2jacinv_numerical(rho_ba, aaxis_ba, num_terms):
  J_ab_inv = np.tile(np.eye(6), (*rho_ba.shape[:-2], 1, 1))

  xi_ba = np.concatenate((rho_ba, aaxis_ba), axis=-2)
  x_small = curlyhat(xi_ba)
  x_small_n = np.tile(np.eye(6), (*rho_ba.shape[:-2], 1, 1))

  bernoulli = np.array([
      1.0, -0.5, 1.0 / 6.0, 0.0, -1.0 / 30.0, 0.0, 1.0 / 42.0, 0.0, -1.0 / 30.0, 0.0, 5.0 / 66.0, 0.0, -691.0 / 2730.0,
      0.0, 7.0 / 6.0, 0.0, -3617.0 / 510.0, 0.0, 43867.0 / 798.0, 0.0, -174611.0 / 330.0
  ])

  for n in range(1, num_terms + 1):
    x_small_n = x_small_n @ (x_small / n)
    J_ab_inv = J_ab_inv + bernoulli[n] * x_small_n

  return J_ab_inv


def _vec2jacinv_rho_aaxis(rho_ba, aaxis_ba, num_terms=0):
  tolerance = 1e-12
  if (npla.norm(aaxis_ba, axis=-2)[0] < tolerance) or (num_terms != 0):
    return _vec2jacinv_numerical(rho_ba, aaxis_ba, num_terms)
  else:
    return _vec2jacinv_analytical(rho_ba, aaxis_ba)


_vec2jacinv_rho_aaxis_vec = np.vectorize(_vec2jacinv_rho_aaxis, signature='(3,1),(3,1),()->(6,6)')


def vec2jacinv(xi_or_rho_ba: np.ndarray, aaxis_ba: Optional[np.ndarray] = None, num_terms: int = 0) -> np.ndarray:
  """Build the 6x6 inverse Jacobian matrix of SE(3).

  For the sake of a notation, we assign subscripts consistence with the transformation,
    J_ab_inverse = J(xi_ba)^{-1},
  Please note that J_ab_inverse is not equivalent to J_ba:
    J(xi_ba)^{-1} != J(-xi_ba)
  For more information see eq. 103 in Barfoot-TRO-2014.

  Args:
    xi_or_rho_ba (np.ndarray): either 6x1 se3 algebra vector xi, or 3x1 translation vector rho when aaxis is None
    aaxis_ba (np.ndarray|None): 3x1 axis angle vector
    num_terms (int): number of terms used in the infinite series; use analytical solution if 0
  Returns:
    np.ndarray: the 6x6 inverse Jacobian matrix of SE(3)
  """
  if aaxis_ba is None:
    assert xi_or_rho_ba.shape[-2:] == (6, 1)
    rho_ba = xi_or_rho_ba[..., :3, :]
    aaxis_ba = xi_or_rho_ba[..., 3:, :]
  else:
    assert xi_or_rho_ba.shape[-2:] == (3, 1)
    assert aaxis_ba.shape[-2:] == (3, 1)
    rho_ba = xi_or_rho_ba

  return _vec2jacinv_rho_aaxis_vec(rho_ba, aaxis_ba, num_terms)


def point2fs(p, scale=None):
  """Turns a homogeneous point into a special 4x6 matrix

  See eq. 72 in Barfoot-TRO-2014 for more information.

  Args:
    p (np.ndarray): a point in homogeneous coordinate if scale is None; otherwise in cartesian coordinate
    scale (float): the scale
  Returns:
    np.ndarray: the special 4x6 matrix
  """
  if scale is None:
    assert p.shape[-2:] == (4, 1)
    scale = p[..., 3, 0]
    p = p[..., :-1, :]
  assert p.shape[-2:] == (3, 1)
  mat = np.zeros(p.shape[:-2] + (4, 6))
  mat[..., :3, :3] = np.array(scale)[..., None, None] * np.tile(np.eye(3), (*p.shape[:-2], 1, 1))
  mat[..., :3, 3:] = -so3op.hat(p)
  return mat


def point2sf(p, scale=None):
  """Turns a homogeneous point into a special 6x4 matrix

  See eq. 72 in Barfoot-TRO-2014 for more information.

  Args:
    p (np.ndarray): a point in homogeneous coordinate if scale is None; otherwise in cartesian coordinate
    scale (float): the scale
  Returns:
    np.ndarray: the special 4x6 matrix
  """
  if scale is None:
    assert p.shape[-2:] == (4, 1)
    scale = p[..., 3, 0]
    p = p[..., :-1, :]
  assert p.shape[-2:] == (3, 1)
  mat = np.zeros(p.shape[:-2] + (6, 4))
  mat[..., 3:, :3] = -so3op.hat(p)
  mat[..., :3, 5:] = p
  return mat



import numpy as np
import numpy.linalg as npla


def hat(x: np.ndarray) -> np.ndarray:
  """Builds the 3x3 skew symmetric matrix

  The hat (^) operator, builds the 3x3 skew symmetric matrix from the 3x1 vector:
    v^ = [0.0  -v3   v2]
         [ v3  0.0  -v1]
         [-v2   v1  0.0]

  See eq. 5 in Barfoot-TRO-2014 for more information.

  Args:
    x (np.ndarray): a 3x1 vector
  Returns:
    np.ndarray: the 3x3 skew symmetric matrix of x
  """
  assert x.shape[-2:] == (3, 1)
  x_hat = np.zeros(x.shape[:-2] + (3, 3))
  x_hat[..., 0, 1] = -x[..., 2, 0]
  x_hat[..., 0, 2] = x[..., 1, 0]
  x_hat[..., 1, 0] = x[..., 2, 0]
  x_hat[..., 1, 2] = -x[..., 0, 0]
  x_hat[..., 2, 0] = -x[..., 1, 0]
  x_hat[..., 2, 1] = x[..., 0, 0]
  return x_hat


def hatinv(x):
  """Gets the 3x1 vector from a 3x3 skew symmetric matrix. Inverse of the hat operator.
  Args:
    x (np.ndarray): a 3x3 skew symmetric matrix
  Returns:
    np.ndarray: the 3x1 vector of x
  """
  assert x.shape[-2:] == (3, 3)
  x_hatinv = np.zeros(x.shape[:-2] + (3, 1))
  x_hatinv[..., 0, 0] = x[..., 2, 1]
  x_hatinv[..., 1, 0] = x[..., 0, 2]
  x_hatinv[..., 2, 0] = x[..., 1, 0]
  return x_hatinv


def validate_rot(C):
  E = np.swapaxes(C, -2, -1) @ C - np.eye(3)
  err = np.max(np.abs(E))
  if err > 1e-7:
    raise RuntimeError('The rotation matrix is not valid. Maximum error: {}'.format(err))


def _vec2rot_analytical(aaxis_ba):

  phi_ba = npla.norm(aaxis_ba, axis=-2, keepdims=True)
  axis = aaxis_ba / phi_ba

  sp = np.sin(phi_ba)
  cp = np.cos(phi_ba)

  return (cp * np.eye(3) + (1 - cp) * (axis @ axis.swapaxes(-1, -2)) + sp * hat(axis))


def _vec2rot_numerical(aaxis_ba, num_terms=10):

  C_ab = np.tile(np.eye(3), (*aaxis_ba.shape[:-2], 1, 1))

  x_small = hat(aaxis_ba)
  x_small_n = np.tile(np.eye(3), (*aaxis_ba.shape[:-2], 1, 1))

  for n in range(1, num_terms + 1):
    x_small_n = x_small_n @ (x_small / n)
    C_ab = C_ab + x_small_n

  return C_ab


def _vec2rot(aaxis_ba, num_terms=0):
  tolerance = 1e-12
  phi_ba = npla.norm(aaxis_ba, axis=-2)[0]
  if (phi_ba < tolerance) or (num_terms != 0):
    return _vec2rot_numerical(aaxis_ba, num_terms)
  else:
    return _vec2rot_analytical(aaxis_ba)


_vec2rot_vec = np.vectorize(_vec2rot, signature='(3,1),()->(3,3)')


def vec2rot(aaxis_ba: np.ndarray, num_terms: int = 0) -> np.ndarray:
  """Builds a rotation matrix using the exponential map.

  This function builds a rotation matrix, C_ab, using the exponential map (from an axis-angle parameterization).
    C_ab = exp(aaxis_ba^),
  where aaxis_ba is a 3x1 axis-angle vector, the magnitude of the angle of rotation can be recovered by finding the norm
  of the vector, and the axis of rotation is the unit-length vector that arises from normalization.
  Note that the angle around the axis, aaxis_ba, is a right-hand-rule (counter-clockwise positive) angle from 'a' to
  'b'.
  Alternatively, we that note that
    C_ba = exp(-aaxis_ba^) = exp(aaxis_ab^).
  Typical robotics convention has some oddity when it comes using this exponential map in practice. For example, if we
  wish to integrate the kinematics:
    d/dt C = omega^ * C,
  where omega is the 3x1 angular velocity, we employ the convention:
    C_20 = exp(deltaTime*-omega^) * C_10,
  Noting that omega is negative (left-hand-rule).
  For more information see eq. 97 in Barfoot-TRO-2014.

  Args:
    aaxis_ba (np.ndarray): the axis-angle vector of the rotation
    num_terms (int): number of terms used in the infinite series approximation of the exponential map
  Returns:
    np.ndarray: the 3x3 rotation matrix of aaxis_ba
  """
  assert aaxis_ba.shape[-2:] == (3, 1)
  return _vec2rot_vec(aaxis_ba, num_terms)


def _rot2vec(C_ab):
  phi_ba = np.arccos(np.clip(0.5 * (np.trace(C_ab) - 1), -1, 1))  # clip to avoid numerical issues
  sinphi_ba = np.sin(phi_ba)

  if np.abs(sinphi_ba) > 1e-9:  # General case: phi_ba is NOT near [0, pi, 2*pi]
    axis = (0.5 / sinphi_ba) * (np.array([C_ab[2, 1] - C_ab[1, 2], C_ab[0, 2] - C_ab[2, 0], C_ab[1, 0] - C_ab[0, 1]
                                         ]).reshape(3, 1))
    return phi_ba * axis

  elif np.abs(phi_ba) > 1e-9:  # phi_ba is near [pi, 2*pi]
    eigval, eigvec = npla.eig(C_ab)
    valid_eigval = np.abs(np.real(eigval) - 1) < 1e-10
    valid_axis = np.real(eigvec[:, valid_eigval])
    axis = valid_axis[:, valid_axis.shape[1] - 1].reshape(3, 1)
    aaxis_ba = phi_ba * axis
    if np.abs(np.trace(_vec2rot(aaxis_ba).T @ C_ab) - 3) > 1e-14:
      aaxis_ba = -aaxis_ba
    return aaxis_ba

  else:
    return np.array([[0., 0., 0.]]).T


_rot2vec_vec = np.vectorize(_rot2vec, signature='(3,3)->(3,1)')


def rot2vec(C_ab: np.ndarray) -> np.ndarray:
  """Computes the matrix log of a rotation matrix

  Compute the inverse of the exponential map (the logarithmic map). This lets us go from a 3x3 rotation matrix back to a
  3x1 axis angle parameterization. In some cases, when the rotation matrix is 'numerically off', this involves some
  'projection' back to SO(3).
    aaxis_ba = ln(C_ab)
  where aaxis_ba is a 3x1 axis angle, where the axis is normalized and the magnitude of the rotation can be recovered by
  finding the norm of the axis angle. Note that the angle around the axis, aaxis_ba, is a right-hand-rule
  (counter-clockwise positive) angle from 'a' to 'b'.
  Alternatively, we that note that
    aaxis_ab = -aaxis_ba = ln(C_ba) = ln(C_ab^T)
  See Barfoot-TRO-2014 Appendix B2 for more information.

  Args:
    C_ab (np.ndarray): a 3x3 rotation matrix
  Returns:
    np.ndarray: the 3x1 axis-angle vector of C_ab
  """
  assert C_ab.shape[-2:] == (3, 3)
  return _rot2vec_vec(C_ab)


def _vec2jac_analytical(aaxis_ba):
  phi_ba = npla.norm(aaxis_ba, axis=-2, keepdims=True)
  axis = aaxis_ba / phi_ba

  sph = np.sin(phi_ba) / phi_ba
  cph = (1 - np.cos(phi_ba)) / phi_ba

  return (sph * np.eye(3) + (1 - sph) * (axis @ np.swapaxes(axis, -1, -2)) + cph * hat(axis))


def _vec2jac_numerical(aaxis_ba, num_terms=10):
  J_ab = np.tile(np.eye(3), (*aaxis_ba.shape[:-2], 1, 1))

  x_small = hat(aaxis_ba)
  x_small_n = np.tile(np.eye(3), (*aaxis_ba.shape[:-2], 1, 1))

  for n in range(1, num_terms + 1):
    x_small_n = x_small_n @ (x_small / (n + 1))
    J_ab = J_ab + x_small_n

  return J_ab


def _vec2jac(aaxis_ba, num_terms=0):
  tolerance = 1e-12
  phi_ba = npla.norm(aaxis_ba, axis=-2)[0]
  if (phi_ba < tolerance) or (num_terms != 0):
    return _vec2jac_numerical(aaxis_ba, num_terms)
  else:
    return _vec2jac_analytical(aaxis_ba)


_vec2jac_vec = np.vectorize(_vec2jac, signature='(3,1),()->(3,3)')


def vec2jac(aaxis_ba: np.ndarray, num_terms: int = 0) -> np.ndarray:
  """Builds the 3x3 Jacobian matrix of SO(3)

  For the sake of a notation, we assign subscripts consistence with the rotation,
    J_ab = J(aaxis_ba),
  although we note to the SO(3) novice that this Jacobian is not a rotation matrix, and should be used with care.
  For more information see eq. 98 in Barfoot-TRO-2014.

  Args:
    aaxis_ba (np.ndarray): a 3x1 axis-angle vector
    num_terms (int): number of terms used in the infinite series approximation of the exponential map
  Returns:
    np.ndarray: the 3x3 Jacobian matrix of SO(3)
  """
  assert aaxis_ba.shape[-2:] == (3, 1)
  return _vec2jac_vec(aaxis_ba, num_terms)


def _vec2jacinv_analytical(aaxis_ba):
  phi_ba = npla.norm(aaxis_ba)
  axis = aaxis_ba / phi_ba
  halfphi = 0.5 * phi_ba

  return (halfphi / np.tan(halfphi) * np.eye(3) + (1 - halfphi / np.tan(halfphi)) * (axis * axis.T) -
          halfphi * hat(axis))


def _vec2jacinv_numerical(aaxis_ba, num_terms=10):
  J_ab_inverse = np.tile(np.eye(3), (*aaxis_ba.shape[:-2], 1, 1))

  x_small = hat(aaxis_ba)
  x_small_n = np.tile(np.eye(3), (*aaxis_ba.shape[:-2], 1, 1))

  bernoulli = np.array([
      1.0, -0.5, 1.0 / 6.0, 0.0, -1.0 / 30.0, 0.0, 1.0 / 42.0, 0.0, -1.0 / 30.0, 0.0, 5.0 / 66.0, 0.0, -691.0 / 2730.0,
      0.0, 7.0 / 6.0, 0.0, -3617.0 / 510.0, 0.0, 43867.0 / 798.0, 0.0, -174611.0 / 330.0
  ])

  for n in range(1, num_terms + 1):
    x_small_n = x_small_n @ (x_small / n)
    J_ab_inverse = J_ab_inverse + bernoulli[n] * x_small_n

  return J_ab_inverse


def _vec2jacinv(aaxis_ba, num_terms=0):
  tolerance = 1e-12
  phi_ba = npla.norm(aaxis_ba, axis=-2)[0]
  if (phi_ba < tolerance) or (num_terms != 0):
    return _vec2jacinv_numerical(aaxis_ba, num_terms)
  else:
    return _vec2jacinv_analytical(aaxis_ba)


_vec2jacinv_vec = np.vectorize(_vec2jacinv, signature='(3,1),()->(3,3)')


def vec2jacinv(aaxis_ba: np.ndarray, num_terms: int = 0) -> np.ndarray:
  """Builds the 3x3 inverse Jacobian matrix of SO(3)

  For the sake of a notation, we assign subscripts consistence with the rotation,
    J_ab_inverse = J(aaxis_ba)^{-1},
  although we note to the SO(3) novice that this Jacobian is not a rotation matrix, and should be used with care. Also,
  please note that J_ab_inverse is not equivalent to J_ba:
    J(aaxis_ba)^{-1} != J(-aaxis_ba)
  For more information see eq. 99 in Barfoot-TRO-2014.

  Args:
    aaxis_ba (np.ndarray): a 3x1 axis-angle vector
    num_terms (int): number of terms used in the infinite series approximation of the exponential map
  Returns:
    np.ndarray: the 3x3 inverse Jacobian matrix of SO(3)
  """
  assert aaxis_ba.shape[-2:] == (3, 1)
  return _vec2jacinv_vec(aaxis_ba, num_terms)


import setuptools

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setuptools.setup(
    name="asrl-pylgmath",
    version="1.0.2",
    author="Yuchen Wu",
    author_email="cheney.wu@mail.utoronto.ca",
    description="Python implementation of lgmath",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/utiasASRL/pylgmath",
    packages=setuptools.find_packages(),
    license="BSD",
    python_requires='>=3.8',
    install_requires=["numpy>=1.21.0", "scipy>=1.7.0"],
)



from __future__ import annotations

from typing import Optional
import numpy as np

from .transformation import Transformation


class TransformationWithCovariance(Transformation):
  """Light weight transformation class with added covariance propagation."""

  def __init__(self,
               *,
               tran_w_cov: Optional[TransformationWithCovariance] = None,
               covariance: Optional[np.ndarray] = None,
               init_cov_to_zero: bool = False,
               **kwargs) -> None:
    """Constructor with several options
    Args:
      tran_w_cov (Optional[TransformationWithCovariance]): an existing instance of this class. (copy construction)
      covariance (np.ndarray): the covariance matrix
      init_cov_to_zero (bool): whether or not to initialize covariance to zero; defaults to False.
    """
    # copy construction
    if tran_w_cov is not None:
      self._covariance = np.copy(tran_w_cov.cov())
      self._covariance_set = tran_w_cov.covariance_set()
      super().__init__(transformation=tran_w_cov)
      return

    super().__init__(**kwargs)

    # construction from a given covariance matrix
    if covariance is not None:
      # TODO assert that the covariance shape is consistent with the underlying transformation
      self._covariance = covariance
      self._covariance_set = True
      return

    # default construction
    self._covariance = np.zeros((6, 6))  # TODO make this the same shape as the underlying transformation
    self._covariance_set = init_cov_to_zero

  def cov(self) -> np.ndarray:
    """Returns a reference to the covariance matrix."""
    if not self._covariance_set:
      raise RuntimeError("Covariance accessed before being set.")
    return self._covariance

  def covariance_set(self) -> bool:
    """Returns whether the covariance has been set."""
    return self._covariance_set

  def set_covariance(self, cov: np.ndarray) -> None:
    """Sets the 6x6 covariance matrix.
    Args:
      cov: the 6x6 covariance matrix
    """
    assert cov.shape[:-2] == self._C_ba.shape[:-2] and cov.shape[-2:] == (6, 6)
    self._covariance = cov
    self._covariance_set = True

  def set_zero_covariance(self) -> None:
    """Sets the covariance to zero."""
    self._covariance = np.zeros_like(*self._C_ba.shape[:-2], 6, 6)
    self._covariance_set = True

  def inverse(self) -> TransformationWithCovariance:
    """Returns the inverse of this transformation including the covariance."""
    temp = TransformationWithCovariance(transformation=super().inverse())
    adjoint_of_inverse = temp.adjoint()
    temp.set_covariance(adjoint_of_inverse @ self._covariance @ adjoint_of_inverse.swapaxes(-2, -1))
    temp.covariance_set = self._covariance_set
    return temp
